import os
import re
from pymeddra.tree import Node, Level


def _create_child_if_not_exists(parent: Node, child_id: str, child_term: str):
    parent_level = parent.level
    # get child level if still appropriate
    child_level = parent_level + 1
    if child_level is None:
        return None
    # get child if exists
    if child_id in parent.children:
        child = parent.children[child_id]
    # create child if not exists
    else:
        child = Node(id=child_id, term=child_term, level=child_level, parent=parent)
        parent.children.update({child_id: child})
    return child


def parse_mdhier(data_dir):
    file = os.path.join(data_dir, "mdhier.asc")
    if not os.path.isfile(file):
        raise FileNotFoundError(f"File {file} not found.")

    # get regex
    id_regex = r"\d{8}"
    term_regex = r"[^$]*"
    regex = rf"({id_regex})\$({id_regex})\$({id_regex})\$({id_regex})\$({term_regex})\$({term_regex})\$({term_regex})\$({term_regex})\$({term_regex})\$\$({id_regex})\$[YN]\$"
    regex = re.compile(regex)

    # get data
    lines = []
    with open(file, "r") as fp:
        lines = fp.readlines()
    text = "\n".join(lines)

    matches = regex.findall(text)

    print(f"Read {len(lines)} lines.")
    if len(lines) != len(matches):
        raise RuntimeError(
            f"Error parsing mdhier.asc. Expected number of matches to match number of lines."
        )

    # create ontology tree
    root = Node(id="root", term="root", level=Level.ROOT)
    for match in matches:
        (
            llt_id,
            hlt_id,
            hltg_id,
            soc_id,
            llt_term,
            hlt_term,
            hltg_term,
            soc_term_1,
            soc_term_2,
            _,
        ) = match
        # recursivelty add the nodes specified by this line
        soc_node = _create_child_if_not_exists(root, soc_id, soc_term_1)
        hltg_node = _create_child_if_not_exists(soc_node, hltg_id, hltg_term)
        hlt_node = _create_child_if_not_exists(hltg_node, hlt_id, hlt_term)
        llt_node = _create_child_if_not_exists(hlt_node, llt_id, llt_term)

    # print(llt_node.parent == hlt_node)
    # print(hlt_node.parent == hltg_node)
    # print(hltg_node.parent == soc_node)
    # print(soc_node.parent == root)

    # print(Level.SOC < Level.LLT)
    # print(Level.SOC > Level.LLT)
    # print(Level.SOC - Level.LLT)

    # print(soc_node.get_parent_at_level(Level.ROOT))
    # print(soc_node.get_parent_at_level(Level.SOC))
    # print(soc_node.get_parent_at_level(Level.LLT))
    # print(llt_node.get_parent_at_level(Level.SOC))

    # print(root.lookup_term("COVID-19 pneumonia"))
    # print(root.lookup_term("COVID-19 pneumonia")[0].parent)
    return root


if __name__ == "__main__":
    data_dir = (
        "/Users/kldooste/Documents/work/pymeddra/data/meddra_23_0_english/MedAscii"
    )
    print("hello parser")
    meddra = parse_mdhier(data_dir)
    meddra.set_lookup_tables()
    print(meddra.lookup_term("COVID-19 pneumonia"))
    print(meddra.lookup_term("covid19 pneumonia"))

    meddra.set_lookup_tables(normalizer=lambda x: x.lower())
    print(meddra.lookup_term("COVID-19 pneumonia"))
    print(meddra.lookup_term("covid-19 pneumonia"))

    nodes1 = meddra.lookup_term("covid-19 pneumonia")
    nodes2 = meddra.lookup_term("Pneumonia measles")

    print(nodes1)
    print(nodes2)

    print(meddra.terms_equivalent("covid-19 pneumonia", "Pneumonia measles")) #true, via 'viral lower respiratory tract infections'
    print(meddra.terms_equivalent("covid-19 pneumonia", "Asymptomatic COVID-19")) #true, via 'coronavirus infections'
    print(meddra.terms_equivalent("covid-19 pneumonia", "pain")) #false



from __future__ import annotations
from pydantic import BaseModel, validator
from typing import Union, Optional, Dict, DefaultDict, List
from enum import Enum
from collections import defaultdict


class Level(Enum):
    ROOT = 0
    SOC = 1
    HLTG = 2
    HLT = 3
    LLT = 4

    def __add__(self, other: int) -> Union[Level, None]:
        if self.ROOT.value <= self.value + other <= self.LLT.value:
            return Level(self.value + other)
        else:
            return None

    def __lt__(self, other: Node) -> bool:
        return self.value < other.value

    def __le__(self, other: Node) -> bool:
        return self.value <= other.value

    def __gt__(self, other: Node) -> bool:
        return self.value > other.value

    def __ge__(self, other: Node) -> bool:
        return self.value >= other.value

    def __sub__(self, other: Node) -> int:
        return self.value - other.value


class Node(BaseModel):
    id: str
    term: str
    level: Level

    parent: Node = None
    children: Dict[str, Node] = {}

    # maps for efficient nested children lookups
    # typically only used for the root node
    id_to_nodes: Optional[DefaultDict[str, List[Node]]] = None
    term_to_nodes: Optional[DefaultDict[str, List[Node]]] = None

    # text normalizer to build lookup tables
    # typically only used for the root node
    normalizer: function = lambda x: x

    def get_parent_at_level(self, target_level: Level) -> Union[Node, None]:
        # check if current node is deeper than target_level
        level_diff = target_level - self.level
        if level_diff > 0:
            return None
        else:
            node = self
            for _ in range(-level_diff):
                node = node.parent
            return node

    def get_children_at_level(self, target_level: Level) -> List[Node]:
        raise NotImplementedError

    def lookup_id(self, id: str) -> Union[List[Node], None]:
        # create lookup tables if they do not exists
        if self.id_to_nodes == None:
            self.set_lookup_tables()
        # lookup

        if id in self.id_to_nodes:
            return self.id_to_nodes[id]
        else:
            return None

    def lookup_term(self, term: str) -> Union[List[Node], None]:
        # create lookup tables if they do not exists
        if self.term_to_nodes == None:
            raise RuntimeError("First set the lookup tables using 'set_lookup_tables'.")
        # lookup
        term = self.normalizer(term)
        if term in self.term_to_nodes:
            return self.term_to_nodes[term]
        else:
            return None

    def set_lookup_tables(self, normalizer: function = lambda x: x) -> None:
        # set normalizer
        self.normalizer = normalizer

        # get a flat list of all nodes
        self.term_to_nodes = defaultdict(list)
        self.id_to_nodes = defaultdict(list)

        # breadth-first traversal
        candidates = [self]
        while len(candidates):
            candidate = candidates.pop(0)
            # track nodes
            self.term_to_nodes[self.normalizer(candidate.term)].append(candidate)
            self.id_to_nodes[candidate.id].append(candidate)
            # add candidate's children to candidate list
            if candidate.children:
                candidates.extend(candidate.children.values())
        return None

    def is_equivalent_node(self, other: Node) -> bool:
        # return true if two nodes are the same, siblings or in a parent-child relation
        return (
            (self == other)
            or (self.parent == other)
            or (self == other.parent)
            or (self.parent == other.parent)
        )

    def terms_equivalent(self, term1: str, term2:str) -> bool:
        # lookup both terms
        term1_nodes = self.lookup_term(term1)
        term2_nodes = self.lookup_term(term2)
        # compare all nodes
        for node1 in term1_nodes:
            for node2 in term2_nodes:
                if node1.is_equivalent_node(node2):
                    return True
        return False

    def __key(self):
        return (self.id, self.term, self.parent, len(self.children))

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other: Node) -> bool:
        # undeep equals operator, does not check children
        return self.__key() == other.__key()

    def __str__(self) -> str:
        return f"Node(id={self.id}, term={self.term}, level={self.level}, parent={self.parent.term if self.parent else None}, #children={len(self.children)})"

    def __repr__(self) -> str:
        return self.__str__()
